/* -*- c++ -*- ***************************************************************/
/* Echtzeitsysteme                                                           */
/*---------------------------------------------------------------------------*/
/*                                                                           */
/*          M U L T I _ L E V E L _ Q U E U E _ S C H E D U L E R _          */
/*                  D E F E R R A B L E _ S E R V E R                        */
/*                                                                           */
/*---------------------------------------------------------------------------*/
/* A multi level queue scheduler that is also capable of dealing with        */
/* deferrable servers.                                                       */
/*****************************************************************************/

#ifndef __abstract_multi_level_queue_scheduler_deferrable_server_h__
#define __abstract_multi_level_queue_scheduler_deferrable_server_h__

#include "infra/types.h"
#include "thread/thread.h"
#include "thread/deferrable_server.h"
#include "thread/multi_level_queue_scheduler.h"
#include "devices/scheduler_implementation_timer.h"
#include "infra/utils.h"
#include "thread/deferrable_server.h"
#include "thread/idle_thread.h"

class Abstract_Multi_Level_Queue_Scheduler_Deferrable_Server
  : public Multi_Level_Queue_Scheduler, 
    public Scheduler_Implementation_Timer			
{

protected:
  ezstubs_uint8 MAX_PRIO;
  Deferrable_Server **ready_servers;

  /**
   * \brief The queue of deferrable servers.
   **/
  Deferrable_Server* first;
  Deferrable_Server* last;

  /**
   * \brief The deferrable server currently running
   **/
  Deferrable_Server* current_server;

  /**
   * \brief The gcd of the replenishmend periods of all deferrable servers.
   **/
  us_type replenishment_gcd;

  /**
   * \brief The time to the next replenishment timer interrupt.
   * \see init()
   **/
  us_type next_replenish;

  /**
   * \brief Do we currently monitor a deferrable server?
   **/
  bool monitor;

  /**
   * \brief The count of timer interrupts.
   **/
  ezstubs_uint32 current_replenishment_tick;
  ezstubs_uint32 replenishment_round;

  /**
   * \brief The maximum priority of all deferrable servers currently ready for
   *        execution.
   **/
  ezstubs_uint8 max_server_prio;

public:

  Abstract_Multi_Level_Queue_Scheduler_Deferrable_Server();

  /**
   * \brief Is a deferrable server ready for execution?
   * \retval true There is a deferrable server ready for execution.
   * \retval false There is no deferrable server ready for execution.
   *
   * A deferrable server is ready for execution if it has been activated and
   * execution budget. There's some deferrable server ready for execution if
   * either the maximum priority among all deferrable servers is greater than
   * zero, or the maximum priority is zero and there is a deferrable server
   * present in the ready queue.
   **/
  bool server_ready();
  
  void update_next_replenish(us_type time);

  /**
   * \brief Init the deferrable servers.
   *
   * All deferrable servers are assigned a replenishment period, the gcd of
   * all replenishment periods is computed. Every gcd time units the timer
   * issues an interrupt and the scheduler keeps track of the interrupt count.
   * For every deferrable server the interrupt count when its budget is 
   * replenished is computed. So at every timer interrupt the queue of 
   * deferrable servers is checked and the budgets of the according servers
   * is replenished.
   **/
  void init(ezstubs_uint8 max_prio);

  /**
   * \brief Replenish the execution budgets of the deferrable servers.
   **/
  void replenish();

  /**
   * \brief Choose the next thread to run,
   * \return The next thread to run.
   **/
  Thread* schedule();

  /**
   * \brief Register a deferrable server with the scheduler
   * \param server The server to be registered with the scheduler
   **/
  void enlist(Deferrable_Server* server);

  /**
   * \brief Activate a deferrable server
   * \param server The server to be activated
   *
   * A deferrable server only is ready for execution when it has been updated
   * and the deferrable server still has execution budget.
   **/
  void add(Deferrable_Server* server);

  void add(Thread* thread);

  /**
   * \brief Finish the execution.
   *
   * The deferrable server finishes execution but its execution budget is not 
   * lost. The deferrable server can be reactivated and will then be executed
   * again. Normal threads are treated as usual.
   **/
  void exit();

  /**
   * \brief Release the processor
   *
   * When a deferrable server releases the remaining execution budget is lost.
   * Normal threads are treated as usual.
   **/
  void yield();

  /**
   * \brief Kill the deferrable server
   * \param server The server to be killed.
   * \see exit()
   *
   * When a deferrable server is killed it preserves is execution budget.
   **/
  void kill(Deferrable_Server* server);
	
  bool isr();
  void dsr();
	
  /**
   * \brief Compute the maximum priority of all deferrable servers that
   *        are currently ready and have some execution budget available.
   **/
  void compute_max_server_prio();

  /**
   * \brief Restart the timer used for replenishment and monitoring
   **/
  void restart_timer(us_type time);

  /**
   * \brief Start the scheduler
   **/
  void start();

  void reschedule();
};

#endif

